﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using SlimDX;

namespace Models.Aqo
{
    using Vertex = Tuple<Vector3, Vector3, Vector3>;

    public class ObjExporter : ImpExpBase
    {
        public static Dictionary<ushort, HashSet<uint>> ValueContains;

        static ObjExporter()
        {
            ValueContains   = new Dictionary<ushort, HashSet<uint>>();
        }

        public ObjExporter(Config config)
            : base(config, true)
        {
        }

        public bool Export(string objfile, AFPFile aqo)
        {
            try
            {
                DoExport(objfile, aqo);

                return true;
            } catch(Exception ex)
            {
                AddMessage("Error: 例外発生({0})", ex.Message);
                return false;
            }
        }

        private void DoExport(string objfile, AFPFile aqo)
        {
            var name    = Path.GetFileNameWithoutExtension(objfile);
          //var dir     = Path.ChangeExtension(objfile, null);
            var dir     = Path.GetDirectoryName(objfile);
            var txtfile = Path.Combine(dir, name+".txt");

            if(dir.Length != 0 && !Directory.Exists(dir))
                Directory.CreateDirectory(dir);

            if(Debug)
                DebugDump(aqo);

            if(null == aqo.VTBF || null == aqo.MESH)
                throw new Exception("未対応の形式です");

            var mtlfile = Path.Combine(dir, name+".mtl");
            var mtlnames= CreateMtlFile(mtlfile, aqo);
            var submeshs= GenerateMeshSet(aqo);

            CreateObjFile(objfile, submeshs);

            if(Debug)
                aqo.Verify();
        }

        private List<string> CreateMtlFile(string mtlfile, AFPFile aqo)
        {
            var dir     = Path.GetDirectoryName(mtlfile);
            var mtlnames= new List<string>();
            var MESH    = aqo.MESH;
            var MATE    = aqo.MATE;
            var TSET    = aqo.TSET;
            var TSTA    = aqo.TSTA;
            var midx    = 0;

            using(var w= new StreamWriter(mtlfile, false, Encoding.UTF8))
            {
                w.WriteLine("# {0}", Path.GetFileName(mtlfile));

                foreach(var i in MESH.Data)
                {
                    var mate    = MATE.Data[i.IndexMATE];
                    var tset    = TSET.Data[i.IndexTSET];
                    var mtlname = string.Format("mtl_{0}_{1}", i.IndexMATE, i.IndexTSET);
                  //var mtlname = string.Format("mtl{0}", midx);

                    mtlnames.Add(mtlname);

                    w.WriteLine("newmtl {0}", mtlname);
                    w.WriteLine("	Ka 0.00000000 0.00000000 0.00000000");
                    w.WriteLine("	Kd 1.00000000 1.00000000 1.00000000");
                    w.WriteLine("	Ks 0.50000000 0.50000000 0.50000000");
                    w.WriteLine("	Ke 0.00000000 0.00000000 0.00000000");

                    uint    tidx;

                    if(0xFFFFFFFF != (tidx= tset._8875[0])) w.WriteLine("	map_Kd   {0}", GetTexFile(TSTA.Data[tidx]._026C, dir));
                  //if(0xFFFFFFFF != (tidx= tset._8875[1])) w.WriteLine("	map_Ks   {0}", GetTexFile(TSTA.Data[tidx]._026C, dir));
                  //if(0xFFFFFFFF != (tidx= tset._8875[2])) w.WriteLine("	map_refl {0}", GetTexFile(TSTA.Data[tidx]._026C, dir));
                    if(0xFFFFFFFF != (tidx= tset._8875[3])) w.WriteLine("	map_Bump {0}", GetTexFile(TSTA.Data[tidx]._026C, dir));

                    ++midx;
                }

                w.Flush();
                w.BaseStream.SetLength(w.BaseStream.Position);
            }

            return mtlnames;
        }

        private void CreateObjFile(string objfile, ExportMeshSet meshes)
        {
            var mtlfile = Path.ChangeExtension(objfile, ".mtl");

            using(var w= new StreamWriter(objfile, false, Encoding.UTF8))
            {
                w.WriteLine("# {0}", Path.GetFileName(objfile));
                w.WriteLine("mtllib {0}", Path.GetFileName(mtlfile));
                w.WriteLine("");

                foreach(var i in meshes.AllMeshes.SelectMany(i => i.Pos.Vertices))
                    w.WriteLine("v  {0:F8} {1:F8} {2:F8}", i.X, i.Y, i.Z);

                w.WriteLine("");

                foreach(var i in meshes.AllMeshes.SelectMany(i => i.Nrm.Vertices))
                    w.WriteLine("vn {0:F8} {1:F8} {2:F8}", i.X, i.Y, i.Z);

                w.WriteLine("");

                foreach(var i in meshes.AllMeshes.SelectMany(i => i.Tex.Vertices))
                    w.WriteLine("vt {0:F8} {1:F8} {2:F8}", i.X, i.Y, i.Z);

                var offpos  = 1;
                var offnrm  = 1;
                var offtex  = 1;

                foreach(var i in meshes.AllMeshes)
                {
                    w.WriteLine("");
                    w.WriteLine("o {0}", i.Name);
                    w.WriteLine("g {0}", i.Name);
                    w.WriteLine("usemtl {0}", i.Material);
                    w.WriteLine("s {0}", 0);

                    for(int j= 0; j < i.Pos.Indices.Count; j+=3)
                    {
                        w.WriteLine("f {0}/{2}/{1} {3}/{5}/{4} {6}/{8}/{7}",
                            i.Pos.Indices[j+0] + offpos, i.Nrm.Indices[j+0] + offnrm, i.Tex.Indices[j+0] + offtex,
                            i.Pos.Indices[j+1] + offpos, i.Nrm.Indices[j+1] + offnrm, i.Tex.Indices[j+1] + offtex,
                            i.Pos.Indices[j+2] + offpos, i.Nrm.Indices[j+2] + offnrm, i.Tex.Indices[j+2] + offtex);
                    }

                    offpos +=i.Pos.Vertices.Count;
                    offnrm +=i.Nrm.Vertices.Count;
                    offtex +=i.Tex.Vertices.Count;
                }
                
                w.Flush();
                w.BaseStream.SetLength(w.BaseStream.Position);
            }
        }

        public ExportMeshSet GenerateMeshSet(AFPFile aqo)
        {
            var midx    = 0;
            var MESH    = aqo.MESH;
            var VSET    = aqo.VSET;
            var PSET    = aqo.PSET;
            var VTXE    = aqo.VTXE;
            var VTXL    = aqo.VTXL;
            var meshes  = new ExportMeshSet();

            foreach(var i in MESH.Data)
            {
                var vset    = VSET.Data[i.IndexVSET];
                var pset    = PSET.Data[i.IndexPSET];
                var vtxe    = VTXE[i.IndexVSET];
                var vtxl    = VTXL[i.IndexVSET];
                var meshname= Mode == 0
                            ? string.Format("mesh_{0}_{1}_{2}_{3}", i.IndexMATE, i.IndexREND, i.IndexSHAD, i.IndexTSET)
                            : string.Format("mesh{0}_{1}_{2}_{3}_{4}", midx, i.IndexMATE, i.IndexREND, i.IndexSHAD, i.IndexTSET);
                var va      = new VertexAccessor(aqo, vset, vtxl, vtxe);
                var indices = pset._86B8;
                var a       = indices[0];
                var b       = indices[1];
                var c       = indices[2];
                var flip    = Reverse;
                var mesh    = meshes.GetOrAdd(meshname);
                mesh.Material= string.Format("mtl_{0}_{1}", i.IndexMATE, i.IndexTSET);

              //for(int j= 0, n= indices.Length; j < n; j+=3)
                for(int j= 2, n= indices.Length; j < n; ++j)    // Triangle strip
                {
                    c   = indices[j];

                    AFPFile.Assert(a < vset.NumVert);
                    AFPFile.Assert(b < vset.NumVert);
                    AFPFile.Assert(c < vset.NumVert);

                    if(a != b && b != c && c != a)
                    {
                        var v0  = GetVertex(va, a);
                        var v1  = GetVertex(va, b);
                        var v2  = GetVertex(va, c);

                        if(flip)    mesh.AddFace(v0, v2, v1);
                        else        mesh.AddFace(v0, v1, v2);
                    }

                    flip= !flip;
                    a   = b;
                    b   = c;
                }

                ++midx;

                if(Mode == 1)
                    meshes.Sweep();
            }

            meshes.Sweep();

            return meshes;
        }

        private Vertex GetVertex(VertexAccessor va, int index)
        {
            var v   = Tuple.Create(
                GetPos.GetPos(new Vector3(va.PX[index], va.PY[index], va.PZ[index])),
                GetNrm.GetNrm(new Vector3(va.NX[index], va.NY[index], va.NZ[index])),
                GetUV.GetUV(new Vector3(va.U0[index], va.V0[index], 0)));

            return v;
        }

        private string GetTexFile(string name, string dir)
        {
            if(dir.Length == 0)
                dir = Environment.CurrentDirectory;

            if(name == "pl_body_diffuse.dds")
                foreach(var i in Directory.GetFiles(dir, "*.dds").Select(i => Path.GetFileName(i)).Where(i => i.StartsWith("pl_")))
                    if(i.IndexOf("_d_") > 0)
                        return i;

            if(name == "pl_body_normal.dds")
                foreach(var i in Directory.GetFiles(dir, "*.dds").Select(i => Path.GetFileName(i)).Where(i => i.StartsWith("pl_")))
                    if(i.IndexOf("_n_") > 0)
                        return i;

            return name;
        }

        public static void DebugDump(AFPFile aqo)
        {
            var name    = Path.GetFileNameWithoutExtension(aqo.FileName);
            var dir     = Path.GetDirectoryName(aqo.FileName);
            var txtfile = Path.Combine(dir, name+".txt");

            // デバッグ用ダンプ
            aqo.Dump(txtfile);

            var dic1    = aqo.VTC0.GroupBy(i => i.SubType).ToDictionary(i => i.Key, i => i.ToArray());

            System.Diagnostics.Debug.Print(aqo.FileName);

            foreach(var i in dic1)
            {
                System.Diagnostics.Debug.Print("  {0}: {1}", i.Key, i.Value.Length);

                var dic2    = i.Value
                                .SelectMany(j => j.DataObj.Params)
                                .GroupBy(j => j.ID)
                                .ToDictionary(j => j.Key, j => j.ToArray());

                foreach(var j in dic2)
                    System.Diagnostics.Debug.Print("    {0:X4}: {1}", j.Key, j.Value.Length);
            }
        }

        public static void AddValueContains(ushort tag, uint value)
        {
            HashSet<uint>   set;

            if(!ValueContains.TryGetValue(tag, out set))
                ValueContains.Add(tag, set= new HashSet<uint>());

            if(!set.Contains(value))
                set.Add(value);
        }

        public static void DumpContains()
        {
            foreach(var i in ValueContains)
            {
                if(i.Value.Count > 32)
                {
                    var min = i.Value.Min();
                    var max = i.Value.Max();
                    
                    System.Diagnostics.Debug.Print("case 0x{0:X4}: CheckRange((uint)j.Value, {1}, {2});                                                      break; // [{3}]",
                        i.Key, min, max, i.Value.Count);
                } else
                {
                    System.Diagnostics.Debug.Print("case 0x{0:X4}: Check     ((uint)j.Value, {1});                                                       break; // [{2}]",
                        i.Key, string.Join(", ", i.Value.Select(j => j.ToString())),  i.Value.Count);
                }
            }
        }

        public static string ToString<T>(T[] a, int count, Func<T, string> func)
        {
            return a.Length > count
                ? string.Format("[{0}]\t{{ {1}, ... }}", a.Length, string.Join(", ", a.Take(count).Select(i => func(i))))
                : string.Format("[{0}]\t{{ {1} }}",      a.Length, string.Join(", ", a.Select(i => func(i))));
        }
    }
}
