﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Text.RegularExpressions;
using Models.Obj;
using SlimDX;

namespace Models.Aqo
{
    public class ObjImporter : ImpExpBase
    {
        public static Regex             RE_ObjName  = new Regex(@"^.*_(\d+)_(\d+)_(\d+)_(\d+)$");

        public ObjImporter(Config config)
            : base(config, false)
        {
        }

        public bool Import(string objfile, AFPFile aqo)
        {
            try
            {
                DoImport(objfile, aqo);

                return true;
            } catch(Exception ex)
            {
                AddMessage("Error: 例外発生({0})", ex.Message);
                return false;
            }
        }

        private void DoImport(string objfile, AFPFile aqo)
        {
            var obj     = ObjFile.FromFile(objfile);
            var submeshes= obj.Meshes.SelectMany(i => i.SubMeshes).ToArray();
            var MESH    = aqo.MESH;
            var VSET    = aqo.VSET;
            var PSET    = aqo.PSET;
            var VTXE    = aqo.VTXE;
            var VTXL    = aqo.VTXL;
            var allmeshes= new List<ImportMesh>();
            var tg      = new TangentCalculator();

            foreach(var i in submeshes)
            {
                if(null == i.Name)
                {
                    AddMessage("Warning: 無名のメッシュを無視しました");
                    continue;
                }

                var m       = RE_ObjName.Match(i.Name);

                if(!m.Success)
                {
                    AddMessage("Warning: 無効な名前のメッシュを無視しました({0})", i.Name);
                    continue;
                }

                var idx_MATE= int.Parse(m.Groups[1].Value);
                var idx_REND= int.Parse(m.Groups[2].Value);
                var idx_SHAD= int.Parse(m.Groups[3].Value);
                var idx_TSET= int.Parse(m.Groups[4].Value);

                // 接線計算
                for(int j= 0; j < i.FaceCount; ++j)
                {
                    var pf  = i.PositionFaces[j];
                    var nf  = i.NormalFaces[j];
                    var tf  = i.TexCoordFaces[j];

                    if(Reverse)
                    {
                        tg.AddFace(
                            GetPos.GetPos(obj.Positions[pf.A]), GetNrm.GetNrm(obj.Normals[nf.A]), GetUV.GetUV(obj.TexCoords[tf.A]),
                            GetPos.GetPos(obj.Positions[pf.C]), GetNrm.GetNrm(obj.Normals[nf.C]), GetUV.GetUV(obj.TexCoords[tf.C]),
                            GetPos.GetPos(obj.Positions[pf.B]), GetNrm.GetNrm(obj.Normals[nf.B]), GetUV.GetUV(obj.TexCoords[tf.B]));
                    } else
                    {
                        tg.AddFace(
                            GetPos.GetPos(obj.Positions[pf.A]), GetNrm.GetNrm(obj.Normals[nf.A]), GetUV.GetUV(obj.TexCoords[tf.A]),
                            GetPos.GetPos(obj.Positions[pf.B]), GetNrm.GetNrm(obj.Normals[nf.B]), GetUV.GetUV(obj.TexCoords[tf.B]),
                            GetPos.GetPos(obj.Positions[pf.C]), GetNrm.GetNrm(obj.Normals[nf.C]), GetUV.GetUV(obj.TexCoords[tf.C]));
                    }
                }

                // 元のスキン情報取得
                var skins   = new ImportSkinSet();

                foreach(var j in MESH.Data.Where(j => j.IndexMATE == idx_MATE
                                                   && j.IndexREND == idx_REND
                                                   && j.IndexSHAD == idx_SHAD
                                                   && j.IndexTSET == idx_TSET))
                {
                    var vset    = VSET.Data[j.IndexVSET];
                    var pset    = PSET.Data[j.IndexPSET];
                    var vtxe    = VTXE[j.IndexVSET];
                    var vtxl    = VTXL[j.IndexVSET];
                    var va      = new VertexAccessor(aqo, vset, vtxl, vtxe);
                    var bones   = null == vset.Bones ? new ushort[] { vset._06BE } : vset.Bones;

                    if(null != va.W0)
                        skins.Create(va, (int)vset.NumVert, bones);
                }

                // メッシュの構築
                var meshes  = new ImportMeshSet();

                for(int j= 0; j < i.PositionFaces.Count; ++j)
                {
                    var pf  = i.PositionFaces[j];
                    var nf  = i.NormalFaces[j];
                    var tf  = i.TexCoordFaces[j];

                    if(Reverse)
                    {
                        meshes.AddFace(
                            skins,
                            GetPos.GetPos(obj.Positions[pf.A]), GetNrm.GetNrm(obj.Normals[nf.A]), GetUV.GetUV(obj.TexCoords[tf.A]),
                            GetPos.GetPos(obj.Positions[pf.C]), GetNrm.GetNrm(obj.Normals[nf.C]), GetUV.GetUV(obj.TexCoords[tf.C]),
                            GetPos.GetPos(obj.Positions[pf.B]), GetNrm.GetNrm(obj.Normals[nf.B]), GetUV.GetUV(obj.TexCoords[tf.B]));
                    } else
                    {
                        meshes.AddFace(
                            skins,
                            GetPos.GetPos(obj.Positions[pf.A]), GetNrm.GetNrm(obj.Normals[nf.A]), GetUV.GetUV(obj.TexCoords[tf.A]),
                            GetPos.GetPos(obj.Positions[pf.B]), GetNrm.GetNrm(obj.Normals[nf.B]), GetUV.GetUV(obj.TexCoords[tf.B]),
                            GetPos.GetPos(obj.Positions[pf.C]), GetNrm.GetNrm(obj.Normals[nf.C]), GetUV.GetUV(obj.TexCoords[tf.C]));
                    }
                }

                foreach(var j in meshes.Meshes)
                {
                    j.IndexMATE = idx_MATE;
                    j.IndexREND = idx_REND;
                    j.IndexSHAD = idx_SHAD;
                    j.IndexTSET = idx_TSET;
                    j.Skinning  = skins.Skins.Count > 0;

                    allmeshes.Add(j);
                }
            }

            tg.Normalize();

            if(0 == allmeshes.Count)
                throw new Exception("インポート可能なメッシュデータがありません");

            // AQOのデータを作成
            var listVTXE= new List<VTXE>();
            var listVTXL= new List<VTXL>();
            var listPSET= new List<PSET.PSETData>();
            var listVSET= new List<VSET.VSETData>();
            var listMESH= new List<MESH.MESHData>();
            int no      = 0;

            foreach(var i in allmeshes)
            {
                var oldMESH = MESH.Data.FirstOrDefault(j => j.IndexMATE == i.IndexMATE
                                                         && j.IndexREND == i.IndexREND
                                                         && j.IndexSHAD == i.IndexSHAD
                                                         && j.IndexTSET == i.IndexTSET);
                var oldVSET = VSET.Data[oldMESH.IndexVSET];
                var oldVTXE = VTXE[oldMESH.IndexVSET];
              //var newVTXE = new VTXE();
                var newVTXE = oldVTXE;  // そのまま使う
                var newVTXL = new VTXL();
                var newPSET = new PSET.PSETData();
                var newVSET = new VSET.VSETData();
                var newMESH = new MESH.MESHData();
                var vbuf    = new byte[i.MeshBuilder.Vertices.Count * oldVSET.VertSize];

                // PSET
                var oldtris     = GeometryUtils.CreateTriangleList(i.MeshBuilder.Indices.Select(j => (ushort)j).ToArray()).ToList();
                var indices     = GeometryUtils.Stripify(i.MeshBuilder.Indices.Select(j => (ushort)j).ToArray());
              //pset.Indices    = mb.Indices.Select(j => (ushort)j).ToArray();
                var newtris     = GeometryUtils.CreateTriangleStrip(indices).ToList();

                if(Debug)
                    GeometryUtils.VerifyTriangleList(oldtris, newtris);

                listPSET.Add(newPSET);
                newPSET.Params.Add(new Param(0x09C6, 8448u));
                newPSET.Params.Add(new Param(0x09BB, 1u));
                newPSET.Params.Add(new Param(0x09BC, (uint)indices.Length));
                newPSET.Params.Add(new Param(0x09B7, (uint)indices.Length));
                newPSET.Params.Add(new Param(0x86B8, indices));
                newPSET.Params.Add(new Param(0x09C5, 0u));

                // VSET
                listVSET.Add(newVSET);
                newVSET.Params.Add(new Param(0x09B6, oldVSET.VertSize));
                newVSET.Params.Add(new Param(0x09BF, oldVSET.NumElem));
                newVSET.Params.Add(new Param(0x09B9, (uint)i.MeshBuilder.Vertices.Count));
                newVSET.Params.Add(new Param(0x09C4, 0u));

                if(i.Skinning)
                {
                    newVSET.Params.Add(new Param(0x08BD, i.Bones.Count));

                    if(i.Bones.Count == 1)
                            newVSET.Params.Add(new Param(0x06BE, (ushort)i.Bones.Keys.First()));
                    else    newVSET.Params.Add(new Param(0x86BE, i.Bones.Keys.Select(k => (ushort)k).ToArray()));
                } else
                {
                    newVSET.Params.Add(new Param(0x08BD, 1));
                  //newVSET.Params.Add(new Param(0x06BE, 0));
                    newVSET.Params.Add(new Param(0x06BE, 1));
                }

                newVSET.Params.Add(new Param(0x09C8, 0u));
                newVSET.Params.Add(new Param(0x09CC, 0u));
                newVSET.Params.Add(new Param(0x09C9, 0u));  // 0x86CAの数
              //newVSET.Params.Add(new Param(0x06CA, (ushort)0));
              //newVSET.Params.Add(new Param(0x86CA, new ushort[0]));

                // VTXE
                listVTXE.Add(newVTXE);  // そのまま使う

                // VTXL
                listVTXL.Add(newVTXL);
                newVTXL.Params.Add(new Param(0x89BA, vbuf));

                // MESH
                listMESH.Add(newMESH);
                newMESH.Params.Add(new Param(0x09B0, oldMESH._09B0));
                newMESH.Params.Add(new Param(0x09C7, oldMESH._09C7));
                newMESH.Params.Add(new Param(0x08B1, i.IndexMATE));
                newMESH.Params.Add(new Param(0x08B2, i.IndexREND));
                newMESH.Params.Add(new Param(0x08B3, i.IndexSHAD));
                newMESH.Params.Add(new Param(0x08B4, i.IndexTSET));
                newMESH.Params.Add(new Param(0x08B5, oldMESH._08B5));
                newMESH.Params.Add(new Param(0x08C0, no));      // IndexVSET
                newMESH.Params.Add(new Param(0x08C1, no));      // IndexPSET
                newMESH.Params.Add(new Param(0x08CD, oldMESH._08CD));   // 0
                newMESH.Params.Add(new Param(0x09C2, oldMESH._09C2));   // 0

                // 頂点データ作成
                var va  = new VertexAccessor(aqo, newVSET, newVTXL, newVTXE);

                for(int j= 0; j < i.MeshBuilder.Vertices.Count; ++j)
                {
                    var pos = i.MeshBuilder.Vertices[j].Item1;
                    var nrm = i.MeshBuilder.Vertices[j].Item2;
                    var uv0 = i.MeshBuilder.Vertices[j].Item3;
                    var uv1 = i.MeshBuilder.Vertices[j].Item4;
                    var bidx= BitConverter.GetBytes(i.MeshBuilder.Vertices[j].Item5);
                    var bwgt= i.MeshBuilder.Vertices[j].Item6;
                    var tan = tg.GetTangent(pos, nrm, uv0);
                    var bin = Vector3.Normalize(Vector3.Cross(nrm, tan));

                    foreach(var k in newVTXE.Data)
                    {
                        switch(k.Usage)
                        {
                        case 0:  va.PX[j]= pos.X;   va.PY[j]= pos.Y;   va.PZ[j]= pos.Z;                      break; // Position
                        case 1:  va.W0[j]= bwgt.X;  va.W1[j]= bwgt.Y;  va.W2[j]= bwgt.Z;  va.W3[j]= bwgt.W;  break; // BlendWeight
                        case 2:  va.NX[j]= nrm.X;   va.NY[j]= nrm.Y;   va.NZ[j]= nrm.Z;                      break; // Normal
                        //case 32: va.TX[j]= tan.X;   va.TY[j]= tan.Y;   va.TZ[j]= tan.Z;                      break; // Tangent
                        //case 33: va.BX[j]= bin.X;   va.BY[j]= bin.Y;   va.BZ[j]= bin.Z;                      break; // Binormal
						case 32: va.TX [j] = 0; va.TY [j] = 0; va.TZ [j] = 0; break; // Tangent
						case 33: va.BX [j] = 0; va.BY [j] = 0; va.BZ [j] = 0; break; // Binormal
                        case 11: va.B0[j]= bidx[0]; va.B1[j]= bidx[1]; va.B2[j]= bidx[2]; va.B3[j]= bidx[3]; break; // BlendIndices
                        case 16: va.U0[j]= uv0.X;   va.V0[j]= uv0.Y;                                         break;
                        case 17: va.U1[j]= uv1.X;   va.V1[j]= uv1.Y;                                         break;
                        default: break;
                        }
                    }
                }
                
                ++no;
            }

            // AQOの構築
            aqo.OBJC.TotalNumVert   = (int)listVSET.Sum(i => i.NumVert);
            aqo.OBJC.TotalNumIndex  = (uint)listPSET.Sum(i => i.NumIndex1);
            aqo.OBJC.NumVSET1       = listVSET.Count;
            aqo.OBJC.NumVSET2       = (uint)listVSET.Count;
            aqo.OBJC.NumVSET3       = (uint)listVSET.Count;
            aqo.OBJC.NumVSET4       = (uint)listVSET.Count;

            var vtc0    = new List<vtc0>();

            vtc0.Add(new vtc0(aqo.ROOT));
            vtc0.Add(new vtc0(aqo.OBJC));
          //vtc0.Add(new vtc0(new VSET(listVSET) { Head= 6 }));
            vtc0.Add(new vtc0(new VSET(listVSET) { Head= (ushort)(listVSET.Count * 2) }));

            #if false
            vtc0.AddRange(listVTXE.Select(i => new vtc0(i)));
            vtc0.AddRange(listVTXL.Select(i => new vtc0(i)));
            #else
            foreach(var i in listVTXE.Zip(listVTXL, Tuple.Create))
            {
                vtc0.Add(new vtc0(i.Item1));
                vtc0.Add(new vtc0(i.Item2));
            }
            #endif

            vtc0.Add(new vtc0(new PSET(listPSET)));
            vtc0.Add(new vtc0(new MESH(listMESH)));
            vtc0.Add(new vtc0(aqo.MATE));
            vtc0.Add(new vtc0(aqo.REND));
            vtc0.Add(new vtc0(aqo.SHAD));
            vtc0.AddRange(aqo.SHAP.Select(i => new vtc0(i)));
            vtc0.Add(new vtc0(aqo.TSTA));
            vtc0.Add(new vtc0(aqo.TSET));
            vtc0.Add(new vtc0(aqo.TEXF));

            aqo.VTC0    = vtc0;

            // 書き出し
          //var aqpfile = Path.ChangeExtension(objfile, ".test.aqp");
            var aqpfile = Path.ChangeExtension(objfile, ".aqp");

            AFPFile.ToFile(aqpfile, aqo);

            if(Debug)
            {
                // 検証
                var chkaqo  = AFPFile.FromFile(aqpfile);

                chkaqo.Dump(Path.ChangeExtension(aqpfile, ".imp.txt"));
                chkaqo.Verify();
            }
        }
    }
}
